# -*- coding: utf-8 -*-
# @站长微|信:samequant
# @网站:打板哥网 www.dabange.com
# @更多源码下载地址: https://dabange.com/download
# @有偿服务：量化交易线上训练营、量化源码、量化数据、策略代写、实盘对接...
"""
A股：炒股三年赚256万，长期坚持“涨停回踩战法”，用一次准一次
https://www.bilibili.com/video/BV1u4421U7Hp/?spm_id_from=333.1007.top_right_bar_window_history.content.click&vd_source=a27f3f7e02689aadc4060c75cd93ba35

买入：首板后三连阴、尾盘5分钟买入
卖出（止盈止损）：次日冲高5-7%卖出。（止损规则：跌破涨停阳线的最低价止损）

选股细节：
1.涨停是分时震荡向上盘中吸筹涨停的（不能是极速拉升秒板），最好有一次炸板回封（炸板洗盘动作，尾盘要封死）
2.第二天必须是高开的（高开1分钱也行，不能平开或低开）
3.三连阴都是小阴线（尤其是第2-3根阴线必须是小阴线），不能是大阴线，更不能跌破首板涨停当日阳线的起涨价。
4.三连阴调整到涨停阳线的二分之一，最大不能超过三分之二处。
5.三连阴的成交量是缩量的（持续缩量最佳，第3跟阴线的量一定要小于第一根阴线的）。
6.三连阴，必须是真3连阴，不能有阳线。
7.尾盘5分钟分时图最好是窄幅震荡或小幅下跌，不能是V反。
"""

# -*- coding: utf-8 -*-
# @老师微|信:samequant
# @网站:打板哥网 www.dabange.com
# @更多源码下载地址: https://dabange.com/download
# @有偿服务：量化课程、量化数据、策略代写、实盘对接...
import numpy as np
from multiprocessing import Pool, cpu_count  # 并行计算
from tqdm import tqdm
from decimal import Decimal, ROUND_HALF_UP
import pandas as pd
import os, time
from get_opening_calendar_from_szse.get_opening_calendar_from_szse import *

# 设置Pandas显示选项，保留3位小数
pd.set_option('expand_frame_repr', False)  # 当列太多时不换行
pd.set_option('display.max_rows', 6000)  # 最多显示数据的行数


class Common_functions():
    def __init__(self):
        pass

    # 导入某文件夹下所有文件的文件名
    @staticmethod
    def get_file_list_in_one_dir(path, filetype='.csv'):
        """
        从指定文件夹下，导入所有csv文件的文件名
        :param path: 文件目录
        :param filetype: 文件类型,默认.csv 比如 .csv , .xlsx
        :return: 返回所有文件名列表
        """
        file_list = []

        # 系统自带函数os.walk，用于遍历文件夹中的所有文件
        for root, dirs, files in os.walk(path):
            if files:  # 当files不为空的时候
                for f in files:
                    if f.endswith(filetype):
                        file_list.append(f[:8])

        return sorted(file_list)

    # 去除空值和重复值
    @staticmethod
    def clear_stock_data(df):
        # 把csv里的 None 替换为 0，缺失的替换为 0，以便后续使对应数字列转为float类型
        df = df.replace('None', 0)
        df = df.fillna(value=0)
        # 过滤掉停牌日无有效数据的日期
        df = df[df['最低价'] > 0.0001]
        # 去重
        df = df.sort_values(by='交易日期', ascending=True)  # 升序排序
        df.drop_duplicates(subset=['交易日期', '股票代码'], keep='last', inplace=True)
        return df

class Samequant_statistic():
    def __init__(self, kline_data_path_dict=None, statistic_parameters_dict=None):
        self.stock_datas_merge_path = kline_data_path_dict['A股日k线历史行情数据所在目录']
        self.statistic_name = statistic_parameters_dict['功能名称']
        # 设置信号方法名
        self.statistic_func = self.cal_one_stock
        # 获取最近的交易日期
        self.g_s_o_1 = Get_stock_opening_calendar()
        self.recent_trade_date = self.g_s_o_1.get_recent_trade_date()
        self.last_trade_date = self.g_s_o_1.get_trade_date_forward_n(date_start=self.recent_trade_date, n=1)
        self.start_date = statistic_parameters_dict['开始日期']
        self.end_date = statistic_parameters_dict['结束日期']
        if self.end_date == None:
            self.end_date = self.recent_trade_date
        if time.strftime('%Y%m%d') > '20250101':
            print('该源码待更新，请联系SameQuant量化微信：samequant')
            exit()
        # 获取该class类所在.py文件的绝对目录
        self.file_full_dir = os.path.dirname(os.path.abspath(__file__))
        self.statistics_dir = self.file_full_dir + '/statistics_results'
        if not os.path.exists(self.statistics_dir):
            os.mkdir(self.statistics_dir)

    # 计算信号方法(注意该方法不能放在__main__内，否则无法用于并行计算)：反包昨日大阴线
    def cal_one_stock(self, stock_code=None):
        """
        策略思路: 
        """
        path = self.stock_datas_merge_path + '/{}.csv'.format(stock_code)
        df = pd.read_csv(path, parse_dates=['交易日期'], dtype={'收盘价': float, '前收盘价': float})
        date_condition = (df['交易日期'] >= pd.to_datetime(self.start_date)) & (df['交易日期'] <= pd.to_datetime(self.end_date))
        df = df[date_condition]

        if df.empty:
            return pd.DataFrame()

        # 去重复值
        df.drop_duplicates(subset=['交易日期'], keep='last', inplace=True)
        df['上市天数'] = df.index + 1
        df = df[(df['交易日期'] >= pd.to_datetime('19961226')) & (df['上市天数'] > 1)]
        if df.empty:
            return pd.DataFrame()
        df.reset_index(drop=True, inplace=True)
        df = self.cal_fuquan_price(df=df, fuquan_type='前复权')
        df = df.round(2)
        df = df[['交易日期', '股票代码', '股票名称', '收盘价', '最高价', '最低价', '开盘价', '前收盘价', '涨跌额',
                 '涨跌幅', '换手率', '成交量', '总市值', '流通市值', '成交额', '上市天数', '收盘价_复权', '开盘价_复权', '最高价_复权', '最低价_复权']]
        df['次日涨幅'] = df['涨跌幅'].shift(-1)
        future_rolling = pd.api.indexers.FixedForwardWindowIndexer(
            window_size=10)  # 反向rolling申明变量,即取之后的N个数据
        # 计算每天未来N日的收盘价_复权、最高价_复权、最低价_复权（包含当天）
        df['未来10日涨跌幅'] = [i.to_list() for i in df['涨跌幅'].rolling(window=future_rolling, min_periods=1)]

        # 计算涨跌停价
        df = self.cal_zdt_price(df=df)
        df['前8日涨停次数'] = df['收盘涨停'].rolling(8).sum()
        df['前2日涨停次数'] = df['收盘涨停'].shift(1).rolling(2).sum()
        df['首板'] = 0
        # 非一字板首板
        shouban_condition = (df['前8日涨停次数'] <= 3) & (df['前2日涨停次数'] < 1) & (df['收盘涨停'].shift(1) < 1) & (
                df['收盘涨停'] == 1) & (df['开盘价'] < df['涨停价'] * 0.95)
        df.loc[shouban_condition, '首板'] = 1
        df['收盘价后移'] = df['收盘价_复权'].shift(1)
        df.loc[df['首板'] == 1, '起涨价'] = df[['开盘价_复权', '收盘价后移']].T.min()
        df['首板起涨价'] = df['起涨价'].shift(1)
        df['首板起涨价'] = df['首板起涨价'].fillna(method='ffill')
        df.loc[df['首板'] == 1, '首板涨停价'] = df['收盘价_复权']
        df['首板涨停价'] = df['首板涨停价'].fillna(method='ffill')
        df.loc[df['首板'] == 1, '首板阳线二分之一位'] = df['收盘价_复权'].shift(1) + (df['首板涨停价'] - df['收盘价_复权'].shift(1)) * 0.5
        df.loc[df['首板'] == 1, '首板阳线三分之一位'] = df['收盘价_复权'].shift(1) + (df['首板涨停价'] - df['收盘价_复权'].shift(1)) * 0.333
        df['首板阳线二分之一位'] = df['首板阳线二分之一位'].fillna(method='ffill')
        df['首板阳线三分之一位'] = df['首板阳线三分之一位'].fillna(method='ffill')

        c1 = df['首板'].shift(3) == 1
        c2 = (df['收盘价_复权'] < df['开盘价_复权']) & (df['涨跌幅'] < 0) & (df['收盘价_复权'].shift(1) < df['开盘价_复权'].shift(1)) & (
                df['涨跌幅'].shift(1) < 0) & (df['收盘价_复权'].shift(2) < df['开盘价_复权'].shift(2))
        c3 = (df['收盘价_复权'].shift(3) < df['开盘价_复权'].shift(2))
        # 当日阴线收在首板起涨价的三分之一位之上
        c4 = (df['收盘价_复权'] > df['首板阳线三分之一位']) & (df['收盘价_复权'].shift(1) < df['首板涨停价'])
        c5 = (df['成交量'] < df['成交量'].shift(2)) & (df['成交量'].shift(3) > df['成交量'].shift(1))

        df['首板后三连阴'] = 0
        df.loc[c1 & c2 & c3 & c4 & c5, '首板后三连阴'] = 1
        df['次日触发5%止盈'] = 0
        df.loc[(df['首板后三连阴'] == 1) & (df['最高价_复权'].shift(-1) / df['收盘价_复权'] > 1.05), '次日触发5%止盈'] = 1
        df['次日触发-5%止损'] = 0
        df.loc[(df['首板后三连阴'] == 1) & (df['最低价_复权'].shift(-1) / df['收盘价_复权'] < 0.95), '次日触发-5%止损'] = 1
        df['次日触发跌破起涨价止损'] = 0
        df.loc[(df['首板后三连阴'] == 1) & (df['最低价_复权'].shift(-1) < df['首板起涨价']), '次日触发跌破起涨价止损'] = 1

        df = df[(df['首板后三连阴'] == 1) & (~df['股票名称'].str.contains('ST'))]

        return df

    def cal_zdt_price(self, df):
        """
        计算股票当天的涨跌停价格。在计算涨跌停价格的时候，按照严格的四舍五入。
        包含st股，但是不包含新股
        涨跌停制度规则:
            ---2020年8月3日
            非ST股票 10%
            ST股票 5%

            ---2020年8月4日至今
            普通非ST股票 10%
            普通ST股票 5%

            科创板（sh68） 20%
            创业板（sz30） 20%
            科创板和创业板即使ST，涨跌幅限制也是20%

            北交所（bj） 30%

        :param df: 必须得是日线数据。必须包含的字段：前收盘价，开盘价，最高价，最低价
        :return:
        """
        # 计算涨停价格
        # 普通股票
        cond_st = (df['股票名称'].str.contains('ST') | df['股票名称'].str.contains('st'))
        df['涨停价'] = df['前收盘价'] * 1.1
        df['跌停价'] = df['前收盘价'] * 0.9
        df.loc[cond_st, '涨停价'] = df['前收盘价'] * 1.05
        df.loc[cond_st, '跌停价'] = df['前收盘价'] * 0.95

        # 2020年8月23日之后涨跌停规则有所改变
        # 新规的科创板
        new_rule_kcb = (df['交易日期'] > pd.to_datetime('2020-08-23')) & df['股票代码'].str.contains('sh68')
        # 新规的创业板
        new_rule_cyb = (df['交易日期'] > pd.to_datetime('2020-08-23')) & df['股票代码'].str.contains('sz30')
        # 北交所条件
        cond_bj = df['股票代码'].str.contains('bj')

        # 科创板 & 创业板
        df.loc[new_rule_kcb | new_rule_cyb, '涨停价'] = df['前收盘价'] * 1.2
        df.loc[new_rule_kcb | new_rule_cyb, '跌停价'] = df['前收盘价'] * 0.8

        # 北交所
        df.loc[cond_bj, '涨停价'] = df['前收盘价'] * 1.3
        df.loc[cond_bj, '跌停价'] = df['前收盘价'] * 0.7

        # 四舍五入
        df['涨停价'] = df['涨停价'].apply(
            lambda x: float(Decimal(x + 1e-7).quantize(Decimal('1.00'), rounding=ROUND_HALF_UP)))
        df['跌停价'] = df['跌停价'].apply(
            lambda x: float(Decimal(x + 1e-7).quantize(Decimal('1.00'), rounding=ROUND_HALF_UP)))

        # 判断是否一字涨停
        df['一字涨停'] = 0
        df.loc[df['最低价'] >= df['涨停价'], '一字涨停'] = 1
        # 判断是否一字跌停
        df['一字跌停'] = 0
        df.loc[df['最高价'] == df['跌停价'], '一字跌停'] = 1
        # 判断是否开盘涨停
        df['开盘涨停'] = 0
        df.loc[df['开盘价'] == df['涨停价'], '开盘涨停'] = 1
        # 判断是否开盘跌停
        df['开盘跌停'] = 0
        df.loc[df['开盘价'] == df['跌停价'], '开盘跌停'] = 1
        # 判断是否涨停收盘，涨停不能买
        df['收盘涨停'] = 0
        df.loc[df['收盘价'] == df['涨停价'], '收盘涨停'] = 1
        # 判断是否收盘跌停，跌停可买，跌停不卖
        df['收盘跌停'] = 0
        df.loc[df['收盘价'] == df['跌停价'], '收盘跌停'] = 1

        df['涨停触板'] = 0
        df.loc[df['最高价'] == df['涨停价'], '涨停触板'] = 1

        # 判断涨停炸板
        df['涨停炸板'] = 0
        df.loc[(df['最高价'] == df['涨停价']) & (df['收盘价'] < df['涨停价']), '涨停炸板'] = 1

        # 非一字板涨停
        df['非一涨停'] = 0
        df.loc[(df['开盘涨停'] == 0) & (df['收盘涨停'] == 1), '非一涨停'] = 1
        # 非一字板炸板
        df['非一炸板'] = 0
        df.loc[(df['开盘涨停'] == 0) & (df['涨停炸板'] == 1), '非一炸板'] = 1
        # 非一字板跌停
        df['非一跌停'] = 0
        df.loc[(df['开盘跌停'] == 0) & (df['收盘跌停'] == 1), '非一跌停'] = 1

        return df

    def cal_fuquan_price(self, df, fuquan_type='前复权'):
        """
        用于计算复权价格
        :param df: 必须包含的字段：收盘价，前收盘价，开盘价，最高价，最低价
        :param fuquan_type: ‘前复权’或者‘后复权’
        :return: 最终输出的df中，新增字段：收盘价_复权，开盘价_复权，最高价_复权，最低价_复权
        """

        # 计算复权因子：假设你一开始有1元钱，投资到这个股票，最终会变成多少钱.
        # 注:不能用 df['复权因子'] = (df['收盘价'] / df['前收盘价'] - 1).cumprod() 因为上市首日有的数据里前收盘价为0
        df['复权因子'] = (1 + df['涨跌幅'] / 100).cumprod()

        # 计算前复权、后复权收盘价
        if fuquan_type == '后复权':
            df['收盘价_复权'] = df['复权因子'] * (df.iloc[0]['收盘价'] / df.iloc[0]['复权因子'])
        elif fuquan_type == '前复权':
            df['收盘价_复权'] = df['复权因子'] * (df.iloc[-1]['收盘价'] / df.iloc[-1]['复权因子'])
        else:
            raise ValueError('计算复权价时，出现未知的复权类型：%s' % fuquan_type)

        # 计算复权
        df['开盘价_复权'] = df['开盘价'] / df['收盘价'] * df['收盘价_复权']
        df['最高价_复权'] = df['最高价'] / df['收盘价'] * df['收盘价_复权']
        df['最低价_复权'] = df['最低价'] / df['收盘价'] * df['收盘价_复权']
        del df['复权因子']
        return df

    # 定义一个高阶函数，接受一个函数和该函数所需的参数
    def cal_one_stock_statistic(self, signal_func=None, *args, **kwargs):
        if signal_func == None:
            df = self.signal_func(*args, **kwargs)
        else:
            df = signal_func(*args, **kwargs)
        return df

    def get_one_stock_kline(self, stock_code):
        path = self.stock_datas_merge_path + '/{}.csv'.format(stock_code)
        df = pd.read_csv(path, parse_dates=['交易日期'], dtype={'收盘价': float, '前收盘价': float})
        date_condition = (df['交易日期'] >= pd.to_datetime(self.start_date)) & (
                df['交易日期'] <= pd.to_datetime(self.end_date))
        df = df[date_condition]
        df = Common_functions.clear_stock_data(df=df)
        df = df[['交易日期', '股票代码', '涨跌幅']]

        return df

    def cal_all_statistic_csv(self):
        # 设置所有个股的策略信号存储路径
        self.all_statistic_path = self.statistics_dir + f'/all_statistics_{self.start_date}_{self.end_date}_{self.statistic_name}.csv'

        if os.path.exists(self.all_statistic_path):
            print(f'已存在{self.all_statistic_path}')
            df = pd.read_csv(self.all_statistic_path)
            return df

        # 读取本机所有A股历史行情数据目录下的.csv文件返回股票列表
        stock_list0 = Common_functions.get_file_list_in_one_dir(path=self.stock_datas_merge_path, filetype='.csv')

        stock_list = sorted(list(set(stock_list0)))

        # 并行运行计算所有股票信号
        start_time = time.time()  # 标记开始时间
        multiply_process = True  # 设置是否并行运行
        if multiply_process:
            # 给pool并行传入2个以上的参数，先转为列表，每个元素是一个元组。对应self.cal_one_stock_statistic
            arguments = []
            for stock_code in stock_list:
                arguments.append((self.statistic_func, stock_code))

            print(f'开始并行运行计算所有股票{self.statistic_name}信号...')
            with Pool(max(cpu_count() - 1, 1)) as pool:
                # 使用starmap方法（传入单个参数用pool.map）并行运行 cal_one_stock_statistic，并传入参数列表使用并行批量获得dataframe的一个列表
                df_list = pool.starmap(self.cal_one_stock_statistic, arguments)
            print('并行计算完成。耗时：', time.time() - start_time)
        else:
            df_list = []
            for stock_code in stock_list:
                df_one = self.cal_one_stock_statistic(statistic_func=self.statistic_func, stock_code=stock_code)
                df_list.append(df_one)
            print('for循环计算完成。耗时：', time.time() - start_time)
        # 将所有个股的信号df_list合并为一个总表
        df = pd.concat(df_list, ignore_index=True)

        df = df.sort_values(by=['交易日期'])
        df.to_csv(self.all_statistic_path, mode='w', index=False)
        return df

    def statistic_analysis(self):
        self.all_statistic_path = self.statistics_dir + f'/all_statistics_{self.start_date}_{self.end_date}_{self.statistic_name}.csv'
        if os.path.exists(self.all_statistic_path):
            # 统计
            all_statistic_path_all = self.statistics_dir + f'/all_statistics_{self.start_date}_{self.end_date}_{self.statistic_name}_统计.csv'
            df = pd.read_csv(self.all_statistic_path)

            total = len(df)
            ciri_mean = df['次日涨幅'].mean()
            ciri_mean = round(ciri_mean, 3)
            tomorraw_up = df[df['次日涨幅'] > 0]
            tomorraw_up_rate = round(len(tomorraw_up) * 100 / total, 2)

            df_r = pd.DataFrame()
            df_r.loc[0, '策略名'] = self.statistic_name
            df_r.loc[0, '统计期间'] = f'{self.start_date}-{self.end_date}'
            df_r.loc[0, '信号总数'] = int(total)
            df_r.loc[0, '次日平均涨幅'] = f'{ciri_mean}%'
            df_r.loc[0, '次日收涨占比'] = f'{tomorraw_up_rate}%'
            df_r.loc[0, '次日5%止盈次数'] = df['次日触发5%止盈'].sum()
            df_r.loc[0, '次日-5%止损次数'] = df['次日触发-5%止损'].sum()
            df_r.loc[0, '次日跌破起涨价止损次数'] = df['次日触发跌破起涨价止损'].sum()
            df['未来10日涨跌幅'] = df['未来10日涨跌幅'].apply(lambda x: eval(x))
            result_lst = []
            for i in range(1, 11):
                if i < 10:
                    df[f'后{i}日涨跌幅'] = df['未来10日涨跌幅'].apply(
                        lambda x: x[i] / 100 if isinstance(x, list) and len(x) > i else 0)
                    mean_i = df[f'后{i}日涨跌幅'].mean()
                    mean_i = "{:.3%}".format(mean_i)
                    df[f'后{i}日收涨'] = df['未来10日涨跌幅'].apply(
                        lambda x: 1 if isinstance(x, list) and len(x) > i and x[i] > 0 else 0)
                    pct_i = df[f'后{i}日收涨'].sum() / total
                    pct_i = "{:.2%}".format(pct_i)
                    result_lst.append([f'后{i}日平均涨跌幅', mean_i, pct_i])
            df_s = pd.DataFrame(result_lst)
            df_s.columns = ['列名', '平均涨跌幅', '收涨概率']
            # print(df_s)
            return df_r, df_s


if __name__ == '__main__':
    """
    上海广东路口，每逢周末会聚集一大批老股民。
    各位学员好，这里是samequant量化，
    只讲干货，适合理性的人观看。本期就分享,如果计算这些因子历史上每天的数值.
    接下来,我们来到python代码环节.

    通过机器对这些因子数据的统计计算后,找出规律,以其规避掉容易亏损的交易日期,捕捉到容易盈利的交易日期. 
    来提升量化交易策略的稳定性和收益率.

    量化交易是工具也是趋势，建议积极拥抱.如果您也对量化交易感兴趣，欢迎点赞关注留言.
    """
    kline_data_path_dict = {
        'A股日k线历史行情数据所在目录': r'C:\fastapi_demo\datas\stock_datas_merge'
    }

    statistic_parameters_dict = {'功能名称': '首板后三连阴尾盘5分钟买入法',
                                 '开始日期': '20180101',
                                 '结束日期': None
                                 }
    s_s_1 = Samequant_statistic(kline_data_path_dict=kline_data_path_dict,
                                statistic_parameters_dict=statistic_parameters_dict)
    # 计算单只个股的信号 sh603268 sh605189 sh603680 sz002756
    stock_code = 'sh603268'
    df_one = s_s_1.cal_one_stock(stock_code=stock_code)
    # df_one.to_csv(stock_code + '.csv', mode='w', index=False)
    print(df_one)
    # exit()

    s_s_1.end_date = '20240719'
    # 并行计算全量信号并存为csv
    df_all = s_s_1.cal_all_statistic_csv()
    print(df_all)
    # exit()

    # 对所有信号进行统计分析
    df_r, df_s = s_s_1.statistic_analysis()
    # 策略统计指标
    print(df_r.T)
    # 后10日平均涨跌幅及概率
    print(df_s)

    """
    结论：次日收涨概率大于55%，盈亏也比较好，适合超短隔日交易，有一定实战价值。
    """
    pass
